from __future__ import annotations
from typing import AnyStr, TYPE_CHECKING, List, Any, Dict

from simple_sql.exceptions import NoSuchColumn, ZeroColumns, ZeroTables

if TYPE_CHECKING:
    from simple_sql.model.column import Column
    from simple_sql.model.table import Table


class Select():
    """
    This class represents a SELECT statement (MySQL compliant).
    Its string form is the syntax needed to execute a SELECT statement
    according to the table. All the parameters are optional to enable 
    the user to specify one clause a time.
    In the future all where clauses will be modeled with a class.

    Parameters
    ----------
    select_clause: Dict[Table, List[Column]], optional  
        The select clause for this select statement.
    from_tables: List[Table], optional
        The from clause for this select statement.
    where_clause: Dict[Table, Dict[Column, Any]], optional
        The where clause for this select statement.

    Attributes
    ----------
    __select_clause: Dict[Table, List[Column]]
        The select clause for this select statement.
    __from_tables: List[Table]
        The from clause for this select statement.
    __where_clause: Dict[Table, Dict[Column, Any]]
        The where clause for this select statement.
    
    Raises
    ------
    NoSuchColumn
        If a column used in either the where or the set clause does
        not exists in the table.
    """


    __from_tables: List[Table] # contains tables for from clause

    __select_clause: Dict[Table, List[Column]] # contains columns for select clause

    __where_clause: Dict[Table, Dict[Column, Any]] # contains a dictionary for the where clause

    def __init__(self, select_clause: Dict[Table, List[Column]] = None, from_tables: List[Table] = None, where_clause: Dict[Table, Dict[Column, Any]] = None):
        self.__select_clause = select_clause
        self.__from_tables = from_tables
        self.__where_clause = where_clause
 
        for table in self.__select_clause.keys(): # check if columns from select clause actually belong to the associated table
            for col in self.__select_clause[table]:
                if not table.contains(col):
                    raise NoSuchColumn('Column {col} (for select clause) does not belong to table {table}'.format(col = col.get_name(), table = table.get_name()))
        
        for table in self.__where_clause.keys(): # check if columns from where clause actually belong to the associated table
            for col in self.__where_clause[table].keys():
                if not table.contains(col):
                    raise NoSuchColumn('Column {col} (for where clause) does not belong to table {table}'.format(col = col.get_name(), table = table.get_name()))


    def __str__(self) -> AnyStr:
        """
        Creates the MySQL compliant string to execute the select statement.

        Returns
        -------
        str
            MySQL compliant string of the select statement execution.
        
        Raises
        ------
        ZeroTables
            If the tables for the FROM clause are not specified.
        """
        # if not self.__select_clause.keys():
        #     raise ZeroColumns('This query is not complete, specify SELECT clause arguments')
        if not self.__from_tables:
            raise ZeroTables('This query is not complete, specify FROM clause arguments')
        string = 'SELECT '
        if not self.__select_clause:
            string += '* '
        else:
            for table in self.__select_clause.keys():
                for col in self.__select_clause[table]:
                    string += '{table}.{column}, '.format(table = table.get_name(), column = col.get_name())
            string = string[:-2] + ' '
        if self.__from_tables:
            string += 'FROM '
            for table in self.__from_tables:
                string += table.get_name() + ', '
        string = string[:-2] + ' '
        if self.__where_clause:
            string += 'WHERE '
            for table in self.__where_clause.keys():
                for col in self.__where_clause[table].keys():
                    if type(self.__where_clause[table][col]) is str:
                        string += '{table}.{column} = "{value}", '.format(table = table.get_name(), column = col.get_name(), value = self.__where_clause[table][col])
                    else:
                        string += '{table}.{column} = {value}, '.format(table = table.get_name(), column = col.get_name(), value = self.__where_clause[table][col])
            string = string[:-2]
        return string



from __future__ import annotations
from typing import TYPE_CHECKING, Any, Dict, List, AnyStr

from simple_sql.exceptions import NoSuchTable, WrongClauseOrder, SyntaxError

if TYPE_CHECKING:
    from simple_sql.model.table import Table
    from simple_sql.model.column import Column
    from simple_sql.metadata import MetaData
from simple_sql.statements.select import Select

class SelectWrapper():
    """
    This class uses the table name, the select clause and where clause
    as strings with the metadata object (handled by ``Database``) to obtain
    the correct ``Table`` object and to create the where and select clauses.
    Using the select and where methods, ``SelectWrapper`` instances will be
    created and returned to add clauses separately.
    In the future all where clauses will be modeled with a class.

    Parameters
    ----------
    metadata: MetaData
        The ``MetaData`` instance of the database.
    select_clause_str: List[AnyStr], optional
        The select clause containing table.column values
        or table values.
    where_clause_str: Dict[AnyStr, Any], optional
        The where clause.

    __metadata: MetaData
        The ``MetaData`` instance of the database.
    __select_clause: Dict[Table, List[Column]]
        The select clause as object.
    __from_tables: List[Table]
        The tables to select from as objects.
    __where_clause: Dict[Table, Dict[Column, Any]]
        The where clause as object.
    __select_clause_str: List[AnyStr]
        The select clause containing table.column values
        or table values.
    __where_clause_str: Dict[AnyStr, Any]
        The where clause as string.
    __table_col_mix: Dict[AnyStr, List[AnyStr]]
        A dictionary of table and list of columns to select
        from that table.

    Raises
    ------
    SyntaxError
        If the select or where clause does not respect specifications.
    """

    __metadata: MetaData

    __select_clause: Dict[Table, List[Column]]
    __from_tables: List[Table]
    __where_clause: Dict[Table, Dict[Column, Any]]

    __select_clause_str: List[AnyStr]
    __where_clause_str: List[AnyStr]

    __table_col_mix: Dict[AnyStr, List[AnyStr]]

    def __init__(self, metadata: MetaData, select_clause_str: List[AnyStr] = None, where_clause_str: Dict[AnyStr, Any] = None):
        self.__metadata = metadata

        self.__select_clause_str = select_clause_str
        self.__where_clause_str = where_clause_str

        self.__from_tables = []
        self.__select_clause = {}
        self.__where_clause = {}
        
        tables: List[AnyStr] = [] # extracting tables and columns from select clause
        columns: Dict[AnyStr, List[AnyStr]] = {}
        if self.__select_clause_str:
            for table_col in self.__select_clause_str:
                parts = table_col.split('.')
                if (len(parts) == 1) and (tables.count(parts[0]) == 0):
                    tables.append(parts[0])
                    for i in self.__metadata.get_table(parts[0]).get_columns():
                        try:
                            columns[parts[0]].append(i.get_name())
                        except KeyError:
                            columns[parts[0]] = []
                            columns[parts[0]].append(i.get_name())
                elif len(parts) == 2:
                    if (tables.count(parts[0]) == 0):
                        tables.append(parts[0])
                    try:
                        columns[parts[0]].append(parts[1])
                    except KeyError:
                        columns[parts[0]] = []
                        columns[parts[0]].append(parts[1])
                else:
                    raise SyntaxError('Select clause must be strings like table or table.column')
            for table in tables:
                self.__from_tables.append(self.__metadata.get_table(table))
                if columns:
                    for col in columns[table]:
                        table_obj = self.__metadata.get_table(table)
                        try:
                            self.__select_clause[table_obj].append(table_obj.get_column(col))
                        except KeyError:
                            self.__select_clause[table_obj] = []
                            self.__select_clause[table_obj].append(table_obj.get_column(col))
        
        self.__table_col_mix = columns

        if self.__where_clause_str: # extracting where object from where string
            if len(self.__from_tables) == 1:
                for where in self.__where_clause_str.keys():
                    parts = where.split('.')
                    if len(parts) == 1:
                        self.__where_clause[self.__from_tables[0]] = {self.__metadata.get_table(tables[0]).get_column(parts[0]): self.__where_clause_str[where]}
                    elif len(parts) == 2:
                        self.__where_clause[self.__from_tables[0]] = {self.__metadata.get_table(tables[0]).get_column(parts[1]): self.__where_clause_str[where]}
                    else:
                        raise SyntaxError('Where clause must be identified as table.column')
            else:
                for where in self.__where_clause_str.keys():
                    parts = where.split('.')
                    if len(parts) == 2:
                        self.__where_clause[self.__metadata.get_table(parts[0])] = {self.__metadata.get_table(parts[0]).get_column(parts[1]): self.__where_clause_str[where]}
                    else:
                        raise SyntaxError('Where clause must be identified as table.column if multiple tables are selected')
    
    
    def get_elements(self) -> Dict[AnyStr, List[AnyStr]]:
        """
        Function used in database to create the dictionary with correct
        table and colum names for better accessibility.

        Returns
        -------
        Dict[AnyStr, List[AnyStr]]
            A dictionary that associates to a table name a list with the
            relative column names.
        """
        return self.__table_col_mix
        
    def __str__(self) -> str:
        """
        Calls the ``Select`` object string representation.

        Returns
        -------
        str
            MySQL compliant string of the select statement execution.
        
        Raises
        ------
        WrongClauseOrder
            If the select clause is not yet specified.
        """
        if self.__select_clause:
            return str(Select(self.__select_clause, self.__from_tables, self.__where_clause))
        else:
            raise WrongClauseOrder('You must specify select clause first')

    # def select(self, select_clause_str: List[AnyStr]) -> SelectWrapper:
    #     return SelectWrapper(self.__metadata, select_clause_str, self.__where_clause_str)

    def where(self, where_clause_str: Dict[AnyStr, Any]) -> SelectWrapper:
        """
        Used to specify the where clause.

        Parameters
        ----------
        where_clause_str: Dict[AnyStr, Any]
            The pairs of column names and values for the where clause.
        
        Returns
        -------
        SelectWrapper
            A ``SelectWrapper`` instance with the same metadata, table
            name and select clause and the where_clause_str passed as parameter.
        
        Raises
        ------
        WrongClauseOrder
            When the where clause is being specified before the select clause.
        """

        if not self.__select_clause_str:
            raise WrongClauseOrder('You have to specify the select clause before')
        return SelectWrapper(self.__metadata, self.__select_clause_str, where_clause_str)


from .delete import Delete
from .drop_table import DropTable
from .insert_into import InsertInto
from .select import Select
from .update import Update

__all__ = (
    'Delete',
    'DropTable',
    'InsertInto',
    'Select',
    'Update',
)


from enum import Enum

class TypesEnum(Enum):
    """
    Enumerative object that contains all the possible types
    MySQL can offer. All the enumeratives are associated with
    a string that represents them in MySQL.

    Attributes
    ----------
    Numeric types:
        TINYINT\n
        SMALLINT\n
        MEDIUMINT\n
        INT\n
        BIGINT\n
        DECIMAL\n
        FLOAT\n
        DOUBLE\n
        BIT\n
    String types:
        CHAR\n
        VARCHAR\n
        BINARY\n
        VARBINARY\n
        TINYBLOB\n
        BLOB\n
        MEDIUMBLOB\n
        LONGBLOB\n
        TINYTEXT\n
        TEXT\n
        MEDIUMTEXT\n
        LONGTEXT\n
        ENUM\n
        SET\n
    Date and time types
        DATE\n
        TIME\n
        DATETIME\n
        TIMESTAMP\n
        YEAR\n
    Spatial data types
        GEOMETRY\n
        POINT\n
        LINESTRING\n
        POLYGON\n
        GEOMETRYCOLLECTION\n
        MULTILINESTRING\n
        MULTIPOINT\n
        MULTIPOLYGON\n
    """
    
    # Numeric types
    TINYINT	= 'TINYINT' # A very small integer
    SMALLINT = 'SMALLINT' # A small integer
    MEDIUMINT = 'MEDIUMINT' # A medium-sized integer
    INT	= 'INT' # A standard integer
    BIGINT = 'BIGINT' # A large integer
    DECIMAL = 'DECIMAL'	# A fixed-point number
    FLOAT = 'FLOAT' # A single-precision floating point number
    DOUBLE = 'DOUBLE' # A double-precision floating point number
    BIT = 'BIT' # A bit field

    # String types
    CHAR = 'CHAR' # A fixed-length nonbinary (character) string
    VARCHAR = 'VARCHAR'	# A variable-length non-binary string
    BINARY = 'BINARY' # A fixed-length binary string
    VARBINARY = 'VARBINARY' # A variable-length binary string
    TINYBLOB = 'TINYBLOB' # A very small BLOB (binary large object)
    BLOB = 'BLOB' # A small BLOB
    MEDIUMBLOB = 'MEDIUMBLOB' # A medium-sized BLOB
    LONGBLOB = 'LONGBLOB' # A large BLOB
    TINYTEXT = 'TINYTEXT' # A very small non-binary string
    TEXT = 'TEXT' # A small non-binary string
    MEDIUMTEXT = 'MEDIUMTEXT' # A medium-sized non-binary string
    LONGTEXT = 'LONGTEXT' # A large non-binary string
    ENUM = 'ENUM' # An enumeration; each column value may be assigned one enumeration member
    SET = 'SET' # A set; each column value may be assigned zero or more SET members

    # Date and time types
    DATE = 'DATE' # A date value in CCYY-MM-DD format
    TIME = 'TIME' # A time value in hh:mm:ss format
    DATETIME = 'DATETIME' # A date and time value inCCYY-MM-DD hh:mm:ssformat
    TIMESTAMP = 'TIMESTAMP' # A timestamp value in CCYY-MM-DD hh:mm:ss format
    YEAR = 'YEAR' # A year value in CCYY or YY format

    # Spatial data types
    GEOMETRY = 'GEOMETRY' # A spatial value of any type
    POINT = 'POINT' # A point (a pair of X-Y coordinates)
    LINESTRING = 'LINESTRING' # A curve (one or more POINT values)
    POLYGON = 'POLYGON' # A polygon
    GEOMETRYCOLLECTION = 'GEOMETRYCOLLECTION' # A collection of GEOMETRY values
    MULTILINESTRING = 'MULTILINESTRING'	# A collection of LINESTRING Values
    MULTIPOINT = 'MULTIPOINT' # A collection of POINTvalues
    MULTIPOLYGON = 'MULTIPOLYGON' # A collection of POLYGONvalues

    




from .delete_wrapper import DeleteWrapper
from .drop_table_wrapper import DropTableWrapper
from .insert_into_wrapper import InsertIntoWrapper
from .select_wrapper import SelectWrapper
from .update_wrapper import UpdateWrapper

__all__ = (
    'DeleteWrapper',
    'DropTableWrapper',
    'InsertIntoWrapper',
    'SelectWrapper',
    'UpdateWrapper',
)


from __future__ import annotations
from typing import TYPE_CHECKING, Any, Dict, AnyStr

from simple_sql.exceptions import ZeroTables, ZeroColumns

if TYPE_CHECKING:
    from simple_sql.model.column import Column
    from simple_sql.model.table import Table


class Update():
    """
    This class represents a UPDATE statement (MySQL compliant).
    Its string form is the syntax needed to execute an UPDATE statement
    according to the table. All the parameters are optional to enable 
    the user to specify one clause a time.
    In the future all where clauses will be modeled with a class.

    Parameters
    ---------
    table: Table, optional
        The table to update.
    set_clause: Dict[Column, Any], optional
        The set clause containg pairs of column and its value.
    where_clause: Dict[Column, Any], optional
        The where clause.

    Attributes
    ----------
    __table: Table
        The table to update.
    __set_clause: Dict[Column, Any]
        The set clause containg pairs of column and value.
    __where_clause: Dict[Column, Any]
        The where clause.
    """

    __table: Table # contains the table on which to perform the update

    __set_clause: Dict[Column, Any] # contains a dictionary of column:value to set

    __where_clause: Dict[Column, Any] # contains a dictionary for the where clause

    def __init__(self, table: Table = None, set_clause: Dict[Column, Any] = None, where_clause: Dict[Column, Any] = None):
        self.__table = table
        self.__set_clause = set_clause
        self.__where_clause = where_clause
        
        # for table in self.__set_clause.keys(): # check if columns from set clause actually belong to the associated table
        #     for col in self.__set_clause[table].keys():
        #         if not table.contains(col):
        #             raise NoSuchColumn('Column {col} (for set clause) does not belong to table {table}'.format(col = col.get_name(), table = table.get_name()))
        
        # for table in self.__where_clause.keys(): # check if columns from where clause actually belong to the associated table
        #     for col in self.__where_clause[table].keys():
        #         if not table.contains(col):
        #             raise NoSuchColumn('Column {col} (for where clause) does not belong to table {table}'.format(col = col.get_name(), table = table.get_name()))

    def __str__(self) -> AnyStr:
        """
        Creates the MySQL compliant string to execute the update statement.

        Returns
        -------
        str
            MySQL compliant string of the update statement execution.

        Raises
        ------
        ZeroTables
            If the table for the UPDATE clause is not specified.
        ZeroColumns
            If the columns for the SET clause are not specified.
        """
        if not self.__table:
            raise ZeroTables('This query is not complete, specify UPDATE clause arguments')
        if not self.__set_clause.keys():
            raise ZeroColumns('This query is not complete, specify SET clause arguments')
        string = 'UPDATE {table} SET '.format(table = self.__table.get_name())
        for col in self.__set_clause.keys():
            if type(self.__set_clause[col]) is str:
                string += '{col} = "{value}", '.format(col = col.get_name(), value = self.__set_clause[col])
            else:
                string += '{col} = {value}, '.format(col = col.get_name(), value = self.__set_clause[col])
        string = string[:-2] + ' '
        if self.__where_clause.keys():
            string += 'WHERE '
            for col in self.__where_clause.keys():
                if type(self.__where_clause[col]) is str:
                    string += '{col} = "{value}", '.format(col = col.get_name(), value = self.__where_clause[col])
                else:
                    string += '{col} = {value}, '.format(col = col.get_name(), value = self.__where_clause[col])
            string = string[:-2]
        return string







class ZeroColumns(Exception):
    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self) -> str:
        if self.message:
            return '{0} '.format(self.message)
        else:
            return ''

class ZeroTables(Exception):
    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self) -> str:
        if self.message:
            return '{0} '.format(self.message)
        else:
            return ''

class SyntaxError(Exception):
    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self) -> str:
        if self.message:
            return '{0} '.format(self.message)
        else:
            return ''

class PrimaryKeyError(Exception):
    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self) -> str:
        if self.message:
            return '{0} '.format(self.message)
        else:
            return ''

class ForeignKeyError(Exception):
    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self) -> str:
        if self.message:
            return '{0} '.format(self.message)
        else:
            return ''

class NoSuchColumn(Exception):
    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self) -> str:
        if self.message:
            return '{0} '.format(self.message)
        else:
            return ''

class WrongClauseOrder(Exception):
    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self) -> str:
        if self.message:
            return '{0} '.format(self.message)
        else:
            return ''
            
class NoSuchTable(Exception):
    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self) -> str:
        if self.message:
            return '{0} '.format(self.message)
        else:
            return ''

class NoSuchDatabase(Exception):
    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self) -> str:
        if self.message:
            return '{0} '.format(self.message)
        else:
            return ''

class DatabaseNotSelected(Exception):
    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self) -> str:
        if self.message:
            return '{0} '.format(self.message)
        else:
            return ''


from .column import Column
from .foreign_key import ForeignKey
from .table import Table
from .type import Type
from .types_enum import TypesEnum

__all__ = (
    'Column',
    'ForeignKey',
    'Table',
    'Type',
    'TypesEnum',
)


from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from simple_sql.model.table import Table


class DropTable():
    """
    This class represents a DROP TABLE statement (MySQL compliant).
    Its string form is the syntax needed to execute a DROP TABLE statement
    according to the table. All the parameters are optional to enable 
    the user to specify one clause a time.

    Parameters
    ----------
    table: Table, optional
        The table to drop.
    
    Attributes
    ----------
    __table: Table
        The table to drop.
    """

    __table: Table

    def __init__(self, table: Table = None) -> None:
        self.__table = table

    def __str__(self) -> str:
        """
        Creates the MySQL compliant string to execute the drop table statement.

        Returns
        -------
        str
            MySQL compliant string of the drop table statement execution.
        """
        return 'DROP TABLE {table}'.format(table = self.__table.get_name())


from __future__ import annotations
from typing import TYPE_CHECKING, List, Any, Dict

from simple_sql.exceptions import SyntaxError

if TYPE_CHECKING:
    from simple_sql.model.table import Table
    from simple_sql.model.column import Column

class InsertInto():
    """
    This class represents a INSERT INTO statement (MySQL compliant).
    Its string form is the syntax needed to execute a INSERT INTO statement
    according to the table. All the parameters are optional to enable 
    the user to specify one clause a time.

    Parameters
    ----------
    table: Table, optional
        The table to insert into.
    columns: List[Column], optional
        To specify a list of columns if you want to insert values for certain columns
    values: List[Dict[Column, Any]], optional
        The values to insert into the table
    
    Attributes
    ----------
    __table: Table
        The table to insert into.
    __columns: List[Column]
        To specify a list of columns if you want to insert values for certain columns
    __values: List[Dict[Column, Any]]
        The values to insert into the table

    Raises
    ------
    SyntaxError
        If the length of value tuples is not valid (with the set of columns
        specified or with the number of columns of the table).
    """

    __table: Table # table in which to insert new values

    __columns: List[Column] # contains the columns that are specified next to the table, optional

    __values: List[Dict[Column, Any]] # contains the values 

    def __init__(self, table: Table = None, columns: List[Column] = None, values: List[Dict[Column, Any]] = None):
        
        self.__table = table

        self.__columns = columns

        self.__values = values

        if columns: # checking that all values entries respect the number of culumns specified
            main = len(self.__columns)
            for i, j in zip(self.__values, range(len(self.__values))):
                if main != len(i.keys()):
                    raise SyntaxError('Number of values provided at row {num} does not match the number of columns specified'.format(num = j))
        else: # checking that all the values entries have the same length
            main = len(self.__table.get_columns())
            for i, j in zip(self.__values, range(len(self.__values))):
                if main != len(i.keys()):
                    raise SyntaxError('Number of values provided at row {num} does not match with the number of columns of the table'.format(num = j))

    def __str__(self) -> str:
        """
        Creates the MySQL compliant string to execute the insert into statement.

        Returns
        -------
        str
            MySQL compliant string of the insert into statement execution.
        """
        string = 'INSERT INTO {table} '.format(table = self.__table.get_name())
        if self.__columns:
            string += '('
            for i in self.__columns:
                string += '{col}, '.format(col = i.get_name())
            string = string[:-2] + ') '
        string += 'VALUES '
        for i in self.__values:
            string += '('
            for j in i.keys():
                if type(i[j]) is str:
                    string += '"{val}", '.format(val = i[j])
                else:
                    string += '{val}, '.format(val = i[j])
            string = string[:-2] + '), '
        string = string[:-2]
        return string



from __future__ import annotations
from typing import Dict, TYPE_CHECKING, AnyStr

if TYPE_CHECKING:
    from simple_sql.model.table import Table
    from simple_sql.metadata import MetaData
    from simple_sql.exceptions import NoSuchTable
from simple_sql.statements.drop_table import DropTable


class DropTableWrapper():
    """
    This class uses the table name a string with the metadata object 
    (handled by ``Database``) to obtain the correct ``Table`` object.

    Parameters
    ----------
    metadata: MetaData
        The ``MetaData`` instance of the database.
    table_str: AnyStr, optional
        The name of the table to update.

    Attributes
    ----------
    __metadata: MetaData
        The ``MetaData`` instance of the database.
    __table: Table
        The ``Table`` object to drop.
    __table_str: AnyStr
        The name of the table to update.
    """

    __metadata: MetaData

    __table: Table

    __table_str: AnyStr

    def __init__(self, metadata: MetaData, table_str: AnyStr = None):
        
        self.__metadata = metadata

        self.__table_str = table_str

    
        self.__table = self.__metadata.get_table(self.__table_str)

    def __str__(self) -> AnyStr:
        """
        Calls the ``DropTable`` object string representation.

        Returns
        -------
        str
            MySQL compliant string of the drop table statement execution.
        """
        return str(DropTable(self.__table))

    # def drop(self, table_str: AnyStr) -> DropTableWrapper:
    #     return DropTableWrapper(self.__metadata, table_str)


from __future__ import annotations
from simple_sql.exceptions import ZeroColumns, PrimaryKeyError, ForeignKeyError, NoSuchColumn
from typing import List, Dict, TYPE_CHECKING
if TYPE_CHECKING:
    from simple_sql.model.foreign_key import ForeignKey
    from simple_sql.metadata import MetaData
from simple_sql.model.column import Column
from simple_sql.model.foreign_key import ForeignKey

# TODO: update primary key constraint creator

class Table():
    """
    Describes a table, the string form of this object is the
    string that declares a table with MySQL compliant syntax.
    Columns, references and primary keys are created using the
    string representation of the column and foreign key objects.
    Inside checks to ensure foreign key compatibility are handled
    here

    Attributes
    ----------
    __columns: List[Column]
        A list of the columns of this table as objects. 
    __name: str
        The name of the table.
    __references: List[ForeignKey]
        A list of the references of this table as objects.
    __if_not_exists: bool
        True if the declaration should include IF NOT EXISTS clause.
    __primary_keys: List[Column]
        A list of the columns that are primary keys.
    __metadata: MetaData
        The ``MetaData`` object relative to this database.

    Parameters
    ----------
    metadata: MetaData
        The ``MetaData`` object relative to this database.
    name: str
        The name of the table.
    \*columns_and_foreign: List[Column or Foreign]
        Contains columns and foreign keys of this table.
    if_not_exists: bool, default True
        True if the table should include IF NOT EXISTS clause.
    
    Raises
    ------
    ZeroColumns
        If no columns are provided to the constructor.
    NoSuchColumn
        If a foreign key is relative to a non-existent column.
    PrimaryKeyError
        If no columns are flagged as primary keys.
    ForeignKeyError
        If a foreign key is relative to a non-existent column.
    """

    __columns: List[Column] = []  # contains a list of the columns of the table

    __name: str  # contains the name of the table

    # contains a list of ForeignKey describing what columns references which column
    __references: List[ForeignKey] = []

    # True if the create table query should cotain IF NOT EXISTS clause, default True
    __if_not_exists: bool

    # list of columns that are primary key
    __primary_keys: List[Column] = []

    __metadata: MetaData

    def __init__(self, metadata: MetaData, name: str, *columns_and_foreign: List, if_not_exists: bool = True):
        
        self.__metadata = metadata

        self.__name = name

        self.__columns = []

        self.__references = []

        self.__primary_keys = []

        if len(columns_and_foreign) == 0:
            raise ZeroColumns('Number of columns must be more than zero')

        for i in columns_and_foreign:
            if type(i) is Column:
                self.__columns.append(i)
            elif type(i) is ForeignKey:
                self.__references.append(i)
            
        temp_from_col: Column        
        check_exists = False # checking that foreigns reference an existent column in this table and setting objects
        for i in self.__references:
            check_exists = False
            for j in self.__columns:
                if i.get_strings()[0] == j.get_name():
                    check_exists = True
                    temp_from_col = j
            if not check_exists:
                raise NoSuchColumn('Foreign key {fk} is invalid, column {col} is not a column of table {table}'.format(fk = repr(i), col = j.get_name(), table = self.get_name()))
            else:
                i.set_objects(from_column=temp_from_col)


        temp_to_table_str: str # checking that foreigns reference an existing table and column
        temp_to_col_str: str
        for i in self.__references:
            temp_to_col_str = i.get_strings()[2]
            temp_to_table_str = i.get_strings()[1]
            temp_to_table = self.__metadata.get_table(temp_to_table_str)
            temp_to_col = temp_to_table.get_column(temp_to_col_str)
            i.set_objects(to_table=temp_to_table, to_column=temp_to_col)

        check_primary = False # checking if there is at least one primary key
        for i in self.__columns:
            if i.is_primary_key():
                self.__primary_keys.append(i)
                check_primary = True
        if not check_primary:
            raise PrimaryKeyError('There should be at least one primary key in table {table}'.format(table = self.__name))
            

        if self.__references: # checking correct assignment of foreign keys
            for i in self.__references:
                if (i.get_from_column() not in self.__columns):
                    raise ForeignKeyError('Foreign key {fkey} does not refer to an existent column'.format(fkey = repr(i)))

        col_dict: Dict[str, Column] = {i.get_name(): i for i in self.__columns}

        self.__dict__.update(col_dict)

        self.__if_not_exists = if_not_exists

    def get_columns(self) -> List[Column]:
        """
        Returns
        -------
        List[Column]
            The columns of the table.
        """
        return self.__columns

    def get_name(self) -> str:
        """
        Returns
        -------
        str
            The name of the table.
        """
        return self.__name

    def get_references(self) -> List[ForeignKey]:
        """
        Returns
        -------
        List[ForeignKey]
            The foreign keys of the table.
        """
        return self.__references

    def get_column(self, column_name: str) -> Column:
        """
        Search and obtain a column in this table given its name.

        Parameters
        ----------
        column_name: str
            The name of the column to search.
        
        Returns
        -------
        Column
            The column named ``column_name``.

        Raises
        ------
        NoSuchColumn
            If there is no column named ``column_name``.
        """
        for i in self.__columns:
            if i.get_name() == column_name:
                return i
        raise NoSuchColumn('Column {col} for table {table} does not exists'.format(col = column_name, table = self.__name))


    def contains(self, column: Column) -> bool:
        """
        Used to check that this table contains a certain column.

        Parameters
        ----------
        column: Column
            The column to search for.
        
        Returns
        -------
        bool
            True if ``column`` is in this table
        """
        return column in self.__columns
        

    def __repr__(self) -> str:
        string = 'Table.{name} '.format(name=self.__name)
        string += str([repr(i) for i in self.__columns]).replace('\'', '') + ', '
        if self.__references:
            string += str([repr(i) for i in self.__references]).replace('\'', '') + ', '
        string += 'IF NOT EXISTS' if self.__if_not_exists else ''
        return '<' + string + '>'

    def __str__(self) -> str:
        """
        Creates the MySQL compliant string to declare the table.

        Returns
        -------
        str
            MySQL compliant string declaration of the table.
        """
        string = 'CREATE TABLE '
        string += 'IF NOT EXISTS ' if self.__if_not_exists else ''
        string += '{name} '.format(name = self.__name)
        string += '('
        for i in self.__columns:
            string += str(i) + ', '
        if self.__references:
            for i in self.__references:
                string += str(i) + ', '
        string += 'PRIMARY KEY ('
        if self.__primary_keys:
            for i in self.__primary_keys:
                string += i.get_name() + ', '
            string = string[:-2] + '))'
        else:
            string = string[:-2] + ')'
        return string


        


from __future__ import annotations
from typing import TYPE_CHECKING

from simple_sql.model.types_enum import TypesEnum


class Type():
    """
    Describes a MySQL type uaing a TypesEnum ad an integer for the length.
    This class is used when declaring columns.

    Attributes
    ----------
    __type: TypesEnum
        The actual type of this ``Type`` instance
    __len: int
        The length of this type, the default value (-1) means that the length
        won't be specified at declaration

    Parameters
    ----------
    type: TypesEnum
        The actual type of this ``Type`` instance
    len: int, default -1
        The length of this type, the default value (-1) means that the length
        won't be specified at declaration
    """

    __type: TypesEnum # contains a TypesEnum object

    __len: int # contains the len that the type is going to be, -1 if not selected

    def __init__(self, type: TypesEnum, len: int = -1):
        self.__type = type
        self.__len = len
    
    def __str__(self) -> str:
        """
        Creates the MySQL compliant string to declare the type.

        Returns
        -------
        str
            MySQL compliant string declaration of the type.
        """
        if self.__type == TypesEnum.VARCHAR:
            string = (self.__type.value + '(10)' if self.__len == -1 else self.__type.value + '({len})'.format(len = self.__len))
        else:
            string = (self.__type.value if self.__len == -1 else self.__type.value + '({len})'.format(len = self.__len))
        return string
    
    def __repr__(self) -> str:
        string = str(self.__type)
        return '<' + string + '>'

    def __eq__(self, o: object) -> bool:
        return (self.__type == o.__type) and (self.__len == o.__len)

    def __hash__(self) -> int:
        return hash((self.__type, self.__len))


from __future__ import annotations
from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    from simple_sql.model.column import Column
    from simple_sql.model.table import Table


class ForeignKey():
    """
    Describes a foreign key, the string form of this object is the
    string that declares a foreign key with MySQL compliant syntax.
    The table of the referencing column is not passed because it is
    handled inside the creation of the relative Table object.

    Attributes
    ----------
    __from_column: Column
        The column in the current table that references another column
    __to_table: Table
        The table to which the column in ``__to_column`` is bound
    __to_column: Column
        The column that is referenced by ``__from_column``
    __from_column_str: str
        The name of ``__from_column``
    __to_table_str: str
        The name of ``__to_table``
    __to_column_str: str
        The name of ``__to_column``
    __on_delete_cascade: bool
        It's True if the reference should cascade on delete

    Parameters
    ----------
    from_column: str
        String name of the column in the current table that references another column
    to_column: str
        String of type table.column, represent the column that is referenced by ``from_column``
    on_delete_cascade: bool, default False
        It's True if the reference should cascade on delete

    Raise
    -----
    SyntaxError
        If the to_column is not identified as table.column
    """

    __from_column: Column  # contains the column object to which the column refers to

    __to_table: Table  # contains the table object to which the column refers to

    __to_column: Column  # contains the column object to which the column refers to

    __from_column_str: str
    __to_column_str: str
    __to_table_str: str

    __on_delete_cascade: bool # True if this foreign key should delete on cascade, False default

    def __init__(self, from_column: str, to_column: str, on_delete_cascade: bool = False):

        if len(to_column.split('.')) != 2:
            raise SyntaxError('The referenced column must be identified as table.column')

        self.__from_column = None
        self.__to_column = None
        self.__to_table = None

        self.__from_column_str = from_column

        self.__to_table_str = to_column.split('.')[0]

        self.__to_column_str = to_column.split('.')[1]

        self.__on_delete_cascade = on_delete_cascade

    def get_from_column(self) -> Column:
        """
        Returns
        -------
        Column
            The referencing column.
        """
        return self.__from_column

    def get_column(self) -> Column:
        """
        Returns
        -------
        Column
            The referenced column.
        """
        return self.__to_column

    def get_table(self) -> Table:
        """
        Returns
        -------
        Table
            The table to which the referenced column belongs to.
        """
        return self.__to_table
    
    def get_strings(self) -> List:
        """
        Returns
        -------
        List
            A list containing the strings respectively of ``from_column``, ``to_table``
            and ``to_column``.
        """
        return [self.__from_column_str, self.__to_table_str, self.__to_column_str]

    def set_objects(self, from_column: Column = None, to_table: Table = None, to_column: Column = None):
        """
        Set the ``from_column``, ``to_table`` and ``to_column`` (all at once or one at a time),
        it is used primarly by the ``Table`` object when parsing the strings.

        Parameters
        ----------
        from_column: Column, optional
            The object that will be assigned to the ``__from_column`` attribute
        to_table: Table, optional
            The object that will be assigned to the ``__to_table`` attribute
        to_column: Column, optional
            The object that will be assigned to the ``__to_column`` attribute
        """
        if not self.__from_column:
            self.__from_column = from_column
        if not self.__to_table:
            self.__to_table = to_table
        if not self.__to_column:
            self.__to_column = to_column

    def __repr__(self) -> str:
        string = 'ForeignKey '
        string += 'ON DELETE CASCADE, ' if self.__on_delete_cascade else ''
        string += '<from {from_col} to {to_col}>'.format(
            from_col=self.__from_column_str, to_col=self.__to_table_str + '.' + self.__to_column_str)
        return '<' + string + '>'

    def __str__(self) -> str:
        """
        Creates the MySQL compliant string to declare the reference.

        Returns
        -------
        str
            MySQL compliant string declaration of the reference.
        """
        string = 'FOREIGN KEY ({from_col}) REFERENCES {table}({column})'.format(
            from_col=self.__from_column_str, table=self.__to_table_str, column=self.__to_column_str)
        return string + (' ON DELETE CASCADE' if self.__on_delete_cascade else '')



from __future__ import annotations
from typing import TYPE_CHECKING, AnyStr, List, Dict, Any

from simple_sql.exceptions import NoSuchTable

if TYPE_CHECKING:
    from simple_sql.model.table import Table
    from simple_sql.metadata import MetaData
    from simple_sql.model.column import Column
from simple_sql.statements.insert_into import InsertInto

class InsertIntoWrapper():
    """
    This class uses the table name, the column list and values list
    as strings with the metadata object (handled by ``Database``) to obtain
    the correct ``Table`` object and to create the column and values list as 
    MySQL compliant clauses.
    Using the insert and values methods, ``InsertIntoWrapper`` instances will be
    created and returned to add clauses separately.

    Parameters
    ----------
    metadata: MetaData
        The ``MetaData`` instance of the database.
    table_str: Table, optional
        The name of the table to insert into.
    columns_str: List[AnyStr], optional
        The list of column names to that defines the set of columns
        to insert the values into.
    values_str: List[Dict[AnyStr, Any]], optional
        The list of pairs of column names and values to insert into the
        table.

    Attributes
    ----------
    __metadata: MetaData
        The ``MetaData`` instance of the database.
    __table: Table
        The ``Table`` object to insert into.
    __columns: List[Column]
        The list of ``Column`` objects that defines the set of columns
        to insert the values into.
    __values: List[Dict[Column, Any]]
        The list of pairs of ``Column`` objects and values to insert into the
        table.
    __table_str: Table
        The name of the table to insert into.
    __columns_str: List[AnyStr]
        The list of column names to that defines the set of columns
        to insert the values into.
    __values_str: List[Dict[AnyStr, Any]]
        The list of pairs of column names and values to insert into the
        table.

    Raises
    ------
    NoSuchTable
        If the table specified does not exist.
    """

    __metadata: MetaData

    __table: Table
    __columns: List[Column]
    __values: List[Dict[Column, Any]]

    __table_str: AnyStr
    __columns_str: List[AnyStr]
    __values_str: List[Dict[AnyStr, Any]] 


    def __init__(self, metadata: MetaData, table_str: AnyStr = None, columns_str: List[AnyStr] = None, values_str: List[Dict[AnyStr, Any]] = None):
        
        self.__metadata = metadata

        self.__table_str = table_str
        self.__columns_str = columns_str
        self.__values_str = values_str

        self.__columns = []
        self.__values = []

        try:
            self.__table = self.__metadata.get_table(self.__table_str)
        except KeyError:
            raise NoSuchTable('Table {table} does not exists'.format(table = self.__table_str))
        
        if self.__columns_str:
            for i in self.__columns_str:
                self.__columns.append(self.__table.get_column(i))
        
        if self.__values_str:
            for i in self.__values_str:
                temp = {}
                for col in i.keys():
                    col_obj = self.__table.get_column(col)
                    val = i[col]
                    temp[col_obj] = val
                self.__values.append(temp)


    def __str__(self) -> str:
        """
        Calls the ``InsertInto`` object string representation.

        Returns
        -------
        str
            MySQL compliant string of the insert into statement execution.
        """
        return str(InsertInto(self.__table, self.__columns, self.__values))

    # def insert_into(self, table_str: AnyStr, columns_str: List[AnyStr]) -> InsertIntoWrapper:
    #     return InsertIntoWrapper(self.__metadata, table_str, columns_str, self.__values_str)

    def values(self, values_str: List[Dict[AnyStr, Any]]) -> InsertIntoWrapper:
        """
        Used to specify the values to insert.

        Parameters
        ----------
        values_str: List[Dict[AnyStr, Any]]
            The list of pairs of column names and values to insert into the
            table.
        
        Returns
        -------
        InsertIntoWrapper
            A ``InsertIntoWrapper`` instance with the same metadata, table
            name and columns and the values_str passed as parameter.
        """
        return InsertIntoWrapper(self.__metadata, self.__table_str, self.__columns_str, values_str)



from __future__ import annotations
from typing import Any, Dict, TYPE_CHECKING, AnyStr

from simple_sql.exceptions import NoSuchTable

if TYPE_CHECKING:
    from simple_sql.model.table import Table
    from simple_sql.metadata import MetaData
    from simple_sql.model.column import Column
from simple_sql.statements.delete import Delete

class DeleteWrapper():
    """
    This class uses the table name and the where clause as strings
    with the metadata object (handled by ``Database``) to obtain
    the correct ``Table`` object and to create the where clause.
    Using the delete and where methods, ``DeleteWrapper`` instances will be
    created and returned to add clauses separately.
    In the future all where clauses will be modeled with a class.

    Parameters
    ----------
    metadata: MetaData, optional
        The ``MetaData`` instance of the database.
    table_str: AnyStr, optional
        The name of the table to delete from.
    where_clause_str: Dict[AnyStr, Any], optional
        The where clause to restrict which entries to delete.

    Attributes
    ----------
    __metadata: MetaData
        The ``MetaData`` instance of the database.
    __table: Table
        The ``Table`` object.
    __where_clause: Dict[Column, Any]
        The where clause generated from the string where clause.
    __table_str: AnyStr
        The name of the table to delete from.
    __where_clause_str: Dict[AnyStr, Any]
        The where clause to restrict which entries to delete.
    """

    __metadata: MetaData

    __table: Table
    __where_clause: Dict[Column, Any]

    __table_str: AnyStr
    __where_clause_str: Dict[AnyStr, Any]

    def __init__(self, metadata: MetaData, table_str: AnyStr, where_clause_str: Dict[AnyStr, Any]):
        
        self.__metadata = metadata

        self.__table_str = table_str
        self.__where_clause_str = where_clause_str

        self.__where_clause = {}

    
        self.__table = self.__metadata.get_table(self.__table_str)
        
        if self.__where_clause_str:
            for i in self.__where_clause_str.keys():
                self.__where_clause[self.__table.get_column(i)] = self.__where_clause_str[i]
        

    def __str__(self) -> str:
        """
        Calls the ``Delete`` object string representation.

        Returns
        -------
        str
            MySQL compliant string of the delete statement execution.
        """
        return str(Delete(self.__table, self.__where_clause))

    # def delete_from(self, table_str: AnyStr) -> DeleteWrapper:
    #     return DeleteWrapper(self.__metadata, table_str, self.__where_clause_str)

    def where(self, where_clause_str: Dict[AnyStr, Any]) -> DeleteWrapper:
        """
        Used to specify the where clause.

        Parameters
        ----------
        where_clause_str: Dict[AnyStr, Any]
            The where clause to restrict which entries to delete.
        
        Returns
        -------
        DeleteWrapper
            A ``DeleteWrapper`` instance with the same metadata and table
            name and the where_clause_str passed as parameter.
        """
        return DeleteWrapper(self.__metadata, self.__table_str, where_clause_str)


from __future__ import annotations

from typing import TYPE_CHECKING, Any

from simple_sql.model.type import Type


class Column():

    """
    Describes a table column, the string form of this object is the
    string that declares a column with MySQL compliant syntax.

    Attributes
    ----------
    __name: str
        The name of the column as a string.
    __type: Type
        The type of the column as a Type object.
    __primary_key: bool
        True if the column is a primary key.
    __unique: bool
        True if the column has UNIQUE keyword.
    __nullable: bool, default True
        True if the column is nullable.
    __constraint: str
        Contains a basic CHECK constraint; *this is a temporary solution,
        the constraints will be handled differently in the future*.
    __default : Any
        Contains the value that will be set as default, if no value
        is passed the DEFAULT clause won't be used; *this is a temporary solution,
        this clause will be handled differently in the future*.
    
    Parameters
    ----------
    name: str
        The name of the column as a string.
    type: Type
        The type of the column as a Type object.
    primary_key: bool, default False
        True if the column is a primary key.
    unique: bool, default False
        True if the column has UNIQUE keyword.
    nullable: bool, default True
        True if the column is nullable.
    constraint: str, optional
        Contains a basic CHECK constraint.
    default : Any, optional
        Contains the value that will be set as default, if no value
        is passed the DEFAULT clause won't be used.
    """

    __name: str  # contains the actual name of the column

    __type: Type  # contains a Type object

    __primary_key: bool  # True if this column is primary key, default False

    __unique: bool  # True if this column has unique restriction, default False

    __nullable: bool  # False if this columns has not null restriction, default True

    # TODO: refactor this, temporary solution
    __constraint: str # Contains a string corresponding the simple constraint
    __default: Any # Contains default value


    def __init__(self, name: str, type: Type, primary_key: bool = False, unique: bool = False, nullable: bool = True, constraint: str = None, default: Any = None):
        self.__name = name
        self.__type = type
        self.__primary_key = primary_key
        self.__unique = unique
        self.__nullable = nullable
        if self.__primary_key:
            self.__unique = False
            self.__nullable = True
        self.__constraint = constraint
        self.__default = default

    def __str__(self) -> str:
        """
        Creates the MySQL compliant string to declare the column.

        Returns
        -------
        str
            MySQL compliant string declaration of the column.
        """
        string = self.__name
        string += ' ' + str(self.__type)
        # string += ' PRIMARY KEY' if self.__primary_key else ''
        string += ' UNIQUE' if self.__unique else ''
        string += '' if self.__nullable else ' NOT NULL'
        string += ' ' + str(self.__constraint) if self.__constraint else ''
        if self.__default:
            if type(self.__default) is str:
                string += ' DEFAULT "' + self.__default + '"'
            else:
                string += ' DEFAULT ' + str(self.__default)
        return string

    def __repr__(self) -> str:
        string = 'Column.{name}, '.format(name=self.__name)
        string += repr(self.__type) + ', '
        string += ' PRIMARY KEY, ' if self.__primary_key else ''
        string += ' UNIQUE, ' if self.__unique else ''
        string += '' if self.__nullable else ' NOT NULL, '
        return '<' + string[:-2] + '>'

    def __eq__(self, o: object) -> bool:
        return (self.__name == o.__name) and \
            (self.__type == o.__type) and \
            (self.__primary_key == o.__primary_key) and \
            (self.__unique == o.__unique) and \
            (self.__nullable == o.__nullable) and \
            (self.__constraint == o.__constraint) and \
            (self.__default == o.__default)

    def is_compatible(self, column: Column) -> bool:
        """
        Defines the way in which two columns are compatible
        in terms of type.

        Parameters
        ----------
        column: Column
            The column to compare with.

        Returns
        -------
        bool
            Returns true if the types of the column is the same
            as the column passed.
        """
        return self.__type == column.__type

    def get_name(self) -> str:
        """
        Returns
        -------
        str
            The name of the column.
        """
        return self.__name

    def get_type(self) -> Type:
        """
        Returns
        -------
        Type
            The type of the column as a Type object.
        """
        return self.__type

    def is_primary_key(self) -> bool:
        """
        Returns
        -------
        bool
            True if the column is primary key.
        """
        return self.__primary_key
    
    def is_nullable(self) -> bool:
        """
        Returns
        -------
        bool
            True if the column is nullable.
        """
        return self.__nullable
    
    def is_unique(self) -> bool:
        """
        Returns
        -------
        bool
            True if the column must be unique.
        """
        return self.__unique

    def get_constraint(self) -> str:
        """
        Returns
        -------
        str
            The string representing the constraint.
        """
        return self.__constraint

    def get_default(self) -> Any:
        """
        Returns
        -------
        Any
            The default value.
        """
        return self.__default

    def __hash__(self) -> int:
        return hash((self.__name, self.__type, self.__primary_key, self.__unique, self.__nullable, self.__constraint))



from __future__ import annotations

from mysql import connector
from mysql.connector.connection import CursorBase, MySQLConnection
from mysql.connector.errors import IntegrityError, ProgrammingError

from simple_sql.metadata import MetaData

from simple_sql.exceptions import DatabaseNotSelected, NoSuchDatabase

from simple_sql.model.table import Table
from simple_sql.model.column import Column
from simple_sql.model.types_enum import TypesEnum
from simple_sql.model.type import Type

from simple_sql.wrappers.select_wrapper import SelectWrapper
from simple_sql.wrappers.delete_wrapper import DeleteWrapper
from simple_sql.wrappers.update_wrapper import UpdateWrapper
from simple_sql.wrappers.drop_table_wrapper import DropTableWrapper
from simple_sql.wrappers.insert_into_wrapper import InsertIntoWrapper

from typing import Dict, AnyStr, List

from pathlib import Path



class Database():
    """
    This is the main class of the module. Using the methods in this class
    you can create, drop, select, update, insert into and delete tables.
    First you must select the database, for now this works only on a
    database level (not serer level like grant options).
    There is also the possibility of executing custom queries
    but you should use that function correctly.
    When initialized the class will connect to the server using the
    provided credentials, create the connection and the cursor.
    
    Note
    ----
    The user must create a :class:`simple_sql.metadata.MetaData` object
    to pass to the instance but the class will take care of updating it.

    Parameters
    ----------
    host: str
        The host of the server to connect to.
    user: str
        The name of the user to use for login.
    password: str
        The string of the password.
    metadata: MetaData
        The metadata instance.
    
    Attributes
    ----------
    __credentials: Dict[str, str]
        The dictionary for the credentials.
    __metadata: MetaData
        The metadata instance.
    __connection: :class:`MySQLConnection`
        The connection to the MySQL server.
    __cursor: :class:`CursorBase`
        The cursor to execute the statements.
    """

    __credentials: Dict[str, str] = {} # dictionary ('host', 'user', 'password')

    __metadata: MetaData # contains metadata for current session of database

    __connection: MySQLConnection

    __cursor: CursorBase

    def __init__(self, host: str, user: str, password: str, metadata: MetaData):

        self.__credentials['host'] = host
        self.__credentials['user'] = user
        self.__credentials['password'] = password

        self.__connection = connector.connect(**self.__credentials)

        self.__metadata = metadata

        self.__cursor = self.__connection.cursor()

    def select_database(self, database: str):
        """
        Use this method to select the working database.

        Parameters
        ----------
        database: str
            The name of the database to select. Use get_databases
            to obtain all the available databases.
        
        Raises
        ------
        NoSuchDatabase
            If the requested database does not exists.
        """
        try:
            self.__cursor.execute('USE {database};'.format(database = database))
            self.__metadata.update_current_database(database)
            self.__connection.commit()
        except ProgrammingError as e:
            if e.errno == 1049:
                raise NoSuchDatabase('Database {db} not selected, try Database.get_databases() to print all available databases'.format(db = database))
        
    def get_databases(self, print_them: bool = True) -> List[str]:
        """
        Use this method to see which databases are available.
        To disable the log print of the list pass False as print_them
        parameter.

        Parameters
        ----------
        print_them: bool, default True
            True if the function should print the list.

        Returns
        -------
        List[str]
            A list of the names of the databases.
        """
        self.__cursor.execute('SHOW DATABASES;')
        fetch = self.__cursor.fetchall()
        databases = []
        for i in fetch:
            databases.append(i[0])
        return databases

    def create_table(self, table: Table):
        """
        Creates a table and adds it to the metadata instance.
        Commits the changes to the connection.

        Parameters
        ----------
        table: Table
            The :class:`Table` to add to the database.
        """
        self.__metadata.add_table(table)

        self.__cursor.execute(str(table))

        self.__connection.commit()

    def select(self, select_clause: List[str]) -> SelectWrapper:
        """
        Initialize a select statement with the select clause.
        The select clause is a list of strings that will be used to
        create the select AND from clause.
        It must be formatted as follows:
        
          * Use the format `table.column` if you want to add a single 
            column to the selection, the table will be added
            to the from clause.
          * Use the format `table` if you want to add all the
            columns of the table to the selection, the table will
            be added to the from clause.

        You can make a list that mixes these two formats.

        Parameters
        ----------
        select_clause: List[str]
            The list of select clause columns.
        
        Returns
        -------
        SelectWrapper
            A :class:`SelectWrapper` instance with the current metadata
            instance and the provided select clause.
        """
        return SelectWrapper(self.__metadata, select_clause_str=select_clause)

    def update(self, table: str) -> UpdateWrapper:
        """
        Initialize a update statement with the name of the table to
        update.

        Parameters
        ----------
        table: str
            The name of the table to update.
        
        Returns
        -------
        UpdateWrapper
            An :class:`UpdateWrapper` instance with the current metadata
            instance and the provided table name.
        """
        return UpdateWrapper(self.__metadata, table_str=table)

    def insert_into(self, table: AnyStr, columns: List[AnyStr] = None) -> InsertIntoWrapper:
        """
        Initialize a insert into statement with the name of the table to
        insert into and eventually the set of column to insert into.

        Parameters
        ----------
        table: str
            The name of the table to insert into.
        columns: List[str]
            The list of column names to insert into.

        Returns
        -------
        InsertIntoWrapper
            An :class:`InsertIntoWrapper` instance with the current
            metadata instance and the provided table name and eventually
            columns list.
        """
        return InsertIntoWrapper(table_str=table, columns_str=columns)
    
    def drop_table(self, table: str) -> DropTableWrapper:
        """
        Initialize a drop table statement with the name of the table
        to drop.
        This method also removes the table from the metadata instance.

        Parameters
        ----------
        table: str
            The name of the table to drop.
        
        Returns
        -------
        DropTableWrapper
            A :class:`DropTableWrapper` instance with the current metadata
            instance and the provided table name.
        """
        self.__metadata.remove_table(table)
        return DropTableWrapper(self.__metadata, table)

    def delete(self, table: str) -> DeleteWrapper:
        """
        Initialize a delete statement with the name of the table
        to delete from.

        Parameters
        ----------
        table: str
            The name of the table to delete from.
        """
        return DeleteWrapper(self.__metadata, table)

    def execute(self, query: SelectWrapper | AnyStr | object) -> List[Dict]: # TODO: change object in anywrapper here and in docstring
        # !!!! Put in anywrapper a get element method interface, all
        # !!!! have to implement it
        """
        Execute a query. To execute the queries you create using the
        appropriate methods you should provide the return value of
        the method. If the query is a select statement you can obtain
        the output of the statement.
        You can also use this method to execute a custom query providing
        a string in MySQL compliant syntax.

        Warning
        -------
        Be careful while executing custom queries because in that case
        the execution is unchecked.

        Parameters
        ----------
        query: str or object
            A string or a wrapper object.
        
        Returns
        -------
        List[Dict] or List[Tuple]
            The result of the fetchall of the execution. If the query is a
            select statement the result is parsed as a list of dictionaries
            with `table.column` key format and the respective value.
            If the query is not a select statement the result is a list
            of tuples.
        """
        if 'SELECT' in str(query): # using dict for select query, find other cases
            self.__cursor.execute(str(query))
            result = self.__create_dictionary(self.__cursor.fetchall(), query.get_elements())
        elif 'ALTER' in str(query):
            print('Feature not yet implemented')
        else:
            self.__cursor.execute(str(query))
            result = self.__cursor.fetchall()
        self.__connection.commit()
        return result

    def __create_dictionary(self, fetchall: List[tuple], elements: Dict[str, List[str]]) -> List[Dict]:
        """
        Private method used to create the dictionary representing
        the result of a select statement using the fetchall and
        the elements.

        Parameters
        ----------
        fetchall: List[Tuple]
            The result of the fetchall of the query execution.
        elements: Dict[str, List[str]]
            A dictionary that associates to a table name a list
            of the table columns' names.

        Returns
        -------
        List[Dict]
            A list of dictionaries with `table.column` as keys and
            values as items.
        """
        result = []
        cols = []
        for item in elements.items():
            for col in item[1]:
                cols.append(str(item[0]) + '.' + str(col))
        for i in fetchall:
            temp = {}
            for col, val in zip(cols, i):
                temp[col] = val
            result.append(temp)
        return result

        


# metadata = MetaData('./simple_sql/metadata_telegram.dump')

# with open(Path('./simple_sql/creds.txt')) as f:
#     temp = f.readlines()

# db = Database(temp[0], 'root', temp[1], metadata)

# db.select_database('telegram')

# test = Table(metadata, 'test', Column('id', Type(TypesEnum.INT), primary_key=True), Column('value', Type(TypesEnum.VARCHAR, 255), default='ASD'))
# print(test)
# db.create_table(test)

# query = db.select(['test.id', 'test.value'])
# print(query)

# print(db.execute(query))




from __future__ import annotations
from typing import TYPE_CHECKING, Any, Dict

from simple_sql.exceptions import NoSuchColumn, ZeroColumns

if TYPE_CHECKING:
    from simple_sql.model.column import Column
    from simple_sql.model.table import Table

class Delete():
    """
    This class represents a DELETE statement (MySQL compliant).
    Its string form is the syntax needed to execute a DELETE statement
    according to the table and where_clause specified. All the parameters
    are optional to enable the user to specify one clause a time.
    In the future all where clauses will be modeled with a class.

    Parameters
    ----------
    table: Table, optional
        The table to delete from.
    where_clause: Dict[Column, Any], optional
        The where clause.
    
    Attributes
    ----------
    __table: Table
        The table to delete from.
    __where_clause: Dict[Column, Any]
        The where clause.
    """

    __table: Table # table to delete from

    __where_clause: Dict[Column, Any] # dict of column and respective value for where clause

    def __init__(self, table: Table = None, where_clause: Dict[Column, Any] = None):
        self.__table = table
        self.__where_clause = where_clause

    def __str__(self) -> str:
        """
        Creates the MySQL compliant string to execute the delete statement.

        Returns
        -------
        str
            MySQL compliant string of the delete statement execution.
        """
        string = 'DELETE FROM {table}'.format(table = self.__table.get_name())
        if self.__where_clause:
            string += ' WHERE '
            for col in self.__where_clause.keys():
                if type(self.__where_clause[col]) is str:
                    string += '{col} = "{val}", '.format(col = col.get_name(), val = self.__where_clause[col])
                else:
                    string += '{col} = {val}, '.format(col = col.get_name(), val = self.__where_clause[col])
        string = string[:-2]
        return string


from .exceptions import (
    ZeroColumns,
    ZeroTables,
    SyntaxError,
    PrimaryKeyError,
    ForeignKeyError,
    NoSuchColumn,
    WrongClauseOrder,
    NoSuchTable,
    NoSuchDatabase,
    DatabaseNotSelected,
)
from .metadata import MetaData
from .database import Database

from .model.table import Table
from .model.column import Column
from .model.foreign_key import ForeignKey
from .model.type import Type
from .model.types_enum import TypesEnum

from .statements.delete import Delete
from .statements.drop_table import DropTable
from .statements.insert_into import InsertInto
from .statements.select import Select
from .statements.update import Update

from .wrappers.delete_wrapper import DeleteWrapper
from .wrappers.drop_table_wrapper import DropTable
from .wrappers.insert_into_wrapper import InsertIntoWrapper
from .wrappers.select_wrapper import Select
from .wrappers.update_wrapper import UpdateWrapper


__all__ = (
    'Database',
    'DatabaseNotSelected',
    'ForeignKeyError',
    'MetaData',
    'NoSuchColumn',
    'NoSuchDatabase',
    'NoSuchTable',
    'PrimaryKeyError',
    'SyntaxError',
    'WrongClauseOrder',
    'ZeroColumns',
    'ZeroTables',
)


from __future__ import annotations
from typing import TYPE_CHECKING, Dict

from simple_sql.exceptions import NoSuchTable

from pathlib import Path
from os.path import isfile
import pickle

if TYPE_CHECKING:
    from simple_sql.model.table import Table

class MetaData():

    """
    This class is an object abstraction of the structure
    of the database, an instance of this class contains all 
    the tables created by USING THIS MODULE, in the future 
    a simple mirror functionality will be added to get tables 
    already declared on the database. 
    This instance also contains the name of the current
    selected database (for now the module only has database-related
    functionalities, not server-related like privileges).
    Every time a change is done on the database structure (new tables, 
    altering actions on tables) the MetaData instance is backed up in
    a file (using python builtin pickle): its path is specified 
    at the creation of the instance. If the backup already
    exists at the specified path then it is restored.

    Parameters
    ----------
    backup_path: str
        The path to the backup file.

    Attributes
    ----------
    __tables: Dict[str, Table]
        A dictionary that pairs a table object with its name.
        Contains all the tables created using ``Database`` or
        restore by a backup.
    __current_database: str
        The name of the current selected database.
    __backup_path: str
        The path of the file to backup to/from. 
    """

    __tables: Dict[str, Table]

    __current_database: str

    __backup_path: str

    def __init__(self, backup_path: str):
        self.__backup_path = backup_path
        if isfile(Path(self.__backup_path)):
            old = self.__restore_backup()
            self.__tables = old.get_tables()
            self.__current_database = old.get_current_database()
        else:
            self.__tables = {}

    def get_table(self, name: str) -> Table:
        """
        Used to obtain the ``Table`` object from table name.

        Parameters
        ----------
        name: str
            The name of the table to search for.
        
        Raises
        ------
        NoSuchTable
            If the table requested is not found.

        Returns
        -------
        Table
            The table that has the requested name.
        """
        try:
            temp = self.__tables[name]
        except KeyError:
            raise NoSuchTable('Table {table} does not exists'.format(table = name))
        return temp

    def get_tables(self) -> Dict[str, Table]:
        """
        Convenience method to obtain the __tables attribute for
        primarly for backup restore.

        Returns
        -------
        Dict[str, Table]
            The mapping of names and relative tables.
        """
        return self.__tables
    
    def add_table(self, table: Table):
        """
        Method used to add a table to this ``MetaData`` instance.
        After adding the table the instance is backed up.

        Parameters
        ----------
        table: Table
            The table object to add.
        """
        if table.get_name() not in self.__tables.keys():
            self.__tables[table.get_name()] = table
        else:
            print('MetaData: Table {table} already exists, updating the existing table'.format(table = table.get_name()))
            self.update_table(table)
        self.__backup()
    
    def remove_table(self, table: str):
        """
        Method used to remove a table from this ``MetaData`` instance.
        After removing the table the instance is backed up.

        Parameters
        ----------
        table: str
            The name of the table to remove.

        Raises
        ------
        NoSuchTable
            If the table name passed does not exists in this
            ``MetaData`` instance.
        """
        try:
            self.__tables.pop(table)
        except KeyError:
            raise NoSuchTable('Table {table} does not exists'.format(table = table))
        self.__backup()

    def update_table(self, table: Table):
        """
        Method used to update a table already present in this 
        ``MetaData`` instance. After updating the table the 
        instance is backed up.

        Parameters
        ----------
        table: Table
            The new table object to replace the old one.

        Raises
        ------
        NoSuchTable
            If the table object passed does not exists in this
            ``MetaData`` instance.
        """
        if table.get_name() in self.__tables.keys():
            self.__tables[table.get_name()] = table
        else:
            raise NoSuchTable('Table {table} does not exists'.format(table = table.get_name()))
        self.__backup()

    def update_current_database(self, database: str):
        """
        Method used to update the current working database.
        The ``MetaData`` instance is backed up updating the
        database.

        Parameters
        ----------
        database: str
            The name of the new database to select.
        """
        self.__current_database = database
        self.__backup()

    def get_current_database(self) -> str:
        """
        Method used to know which is the current working
        database.
        
        Returns
        -------
        str
            The name of the database.
        """
        return self.__current_database

    def __backup(self):
        """
        Private method that dumps (using pickle) the current state
        of this :class:`simple_sql.metadata.MetaData` instance at the specified backup_path.
        """
        with open(Path(self.__backup_path), 'wb+') as f:
            pickle.dump(self, f)

    def __restore_backup(self) -> MetaData:
        """
        Private method used to load a backup of a ``MetaData`` instance
        from the file at the backup_path of this instance.

        Returns
        -------
        MetaData
            The backed up metadata instace.
        """
        with open(Path(self.__backup_path), 'rb') as f:
            return pickle.load(f)


from __future__ import annotations
from typing import TYPE_CHECKING, Any, Dict, AnyStr

from simple_sql.exceptions import NoSuchTable, WrongClauseOrder

if TYPE_CHECKING:
    from simple_sql.model.table import Table
    from simple_sql.model.column import Column
    from simple_sql.metadata import MetaData
from simple_sql.statements.update import Update

class UpdateWrapper():

    """
    This class uses the table name, the set clause and where clause
    as strings with the metadata object (handled by ``Database``) to obtain
    the correct ``Table`` object and to create the where and set clauses.
    Using the update, set and where methods, ``UpdateWrapper`` instances will be
    created and returned to add clauses separately.
    In the future all where clauses will be modeled with a class.

    Parameters
    ----------
    metadata: MetaData, optional
        The ``MetaData`` instance of the database.
    table_str: AnyStr, optional
        The name of the table to update.
    set_clause_str: Dict[AnyStr, Str], optional
        The set clause containing which column to update
        and with which value.
    where_clause_str: Dict[AnyStr, Str], optional
        The where clause to restrict which entries to updates.

    Attributes
    ----------
    __metadata: MetaData, optional
        The ``MetaData`` instance of the database.
    __table: Table
        The ``Table`` object to update.
    __set_clause: Dict[AnyStr, Any]
        The set clause generated from the string set clause.
    __where_clause: Dict[AnyStr, Any]
        The where clause generated from the string where clause.
    __table_str: AnyStr, optional
        The name of the table to update.
    __set_clause_str: Dict[AnyStr, Str], optional
        The set clause containing which column to update
        and with which value.
    __where_clause_str: Dict[AnyStr, Str], optional
        The where clause to restrict which entries to updates.

    Raises
    ------
    NoSuchTable
        If the table specified does not exists.
    """

    __metadata: MetaData

    __table: Table
    __set_clause: Dict[Column, Any]
    __where_clause: Dict[Column, Any]

    __table_str: AnyStr
    __set_clause_str: Dict[AnyStr, Any]
    __where_clause_str: Dict[AnyStr, Any]


    def __init__(self, metadata: MetaData, table_str: AnyStr = None, set_clause_str: Dict[AnyStr, Any] = None, where_clause_str: Dict[AnyStr, Any] = None):
        self.__metadata = metadata

        self.__table_str = table_str
        self.__set_clause_str = set_clause_str
        self.__where_clause_str = where_clause_str

        self.__set_clause = {}
        self.__where_clause = {}

        try:
            self.__table = self.__metadata.get_table(self.__table_str)
        except KeyError:
            raise NoSuchTable('Table {table} does not exists'.format(table = self.__table_str))

        for col in self.__where_clause_str.keys():
            col_obj = self.__table.get_column(col)
            self.__where_clause[col_obj] = self.__where_clause_str[col]

        for col in self.__set_clause_str.keys():
            col_obj = self.__table.get_column(col)
            self.__set_clause[col_obj] = self.__set_clause_str[col]
    
    def __str__(self) -> str:
        """
        Calls the ``Update`` object string representation.

        Returns
        -------
        str
            MySQL compliant string of the update statement execution.
        """
        return str(Update(self.__table, self.__set_clause, self.__where_clause))

    # def update(self, table_str: AnyStr) -> UpdateWrapper:
    #     return UpdateWrapper(self.__metadata, table_str, self.__set_clause_str, self.__where_clause_str)

    def set(self, set_clause_str: Dict[AnyStr, Any]) -> UpdateWrapper:
        """
        Used to specify the set clause.

        Parameters
        ----------
        set_clause_str: Dict[AnyStr, Any]
            The set clause containing which column to update
            and with which value.
        
        Returns
        -------
        UpdateWrapper
            A ``UpdateWrapper`` instance with the same metadata, table
            name and where clause and the set_clause_str passed as parameter.
        """
        return UpdateWrapper(self.__metadata, self.__table_str, set_clause_str, self.__where_clause_str)

    def where(self, where_clause_str: Dict[AnyStr, Any]) -> UpdateWrapper:
        """
        Used to specify the where clause.

        Parameters
        ----------
        where_clause_str: Dict[AnyStr, Any]
            The where clause to restrict which entries to update.
        
        Returns
        -------
        UpdateWrapper
            A ``UpdateWrapper`` instance with the same metadata, table
            name and set clause and the where_clause_str passed as parameter.

        Raises
        ------
        WrongClauseOrder
            If the where clause is being specified before the set clause.
        """
        if not self.__set_clause:
            raise WrongClauseOrder('You have to specify the set clause before')
        return UpdateWrapper(self.__metadata, self.__table_str, self.__set_clause_str, where_clause_str)
    